"""Helper functions for streams."""

from typing import Literal, Dict

import pandas as pd
from ml4pd import registry
from ml4pd.streams import MaterialStream
from pydantic import confloat, validate_arguments


@validate_arguments
def filter_by_mol_perc(
    output_stream: MaterialStream,
    mol_perc: confloat(ge=0, le=1),
    column_name: str = None,
    input_stream: MaterialStream = None,
    rel_perc: confloat(ge=0, le=1) = 0,
    mol_perc_direction: Literal["ge", "le"] = "ge",
    rel_perc_direction: Literal["ge", "le"] = "ge",
) -> pd.Index:
    """A function to pick instances of output stream that pass a specified threshold.

    Args:
        output_stream (MaterialStream): stream to filter.
        mol_perc (float): Only pick instances with output/sum(output) >= mol_perc for any
            column unless column_name is specified.
        column_name (str, optional): Column name to focus on filtering. Defaults to None.
        input_stream (MaterialStream, optional): To further filter results. Defaults to None.
        rel_perc (float, optional): Only instances if output/input >= rel_perc. Defaults to 0.

    Returns:
        pd.Index: Indices of instances that pass filter(s).
    """

    out_flow = output_stream.flow
    flow_perc = out_flow.div(out_flow.sum(axis=1), axis=0)

    if input_stream is not None:
        inp_flow = input_stream.flow
        flow_perc_rel_to_inp = out_flow.div(inp_flow)

    if column_name is not None:

        if mol_perc_direction == "ge":
            index_1 = set(out_flow[flow_perc[column_name] >= mol_perc].index)
        else:
            index_1 = set(out_flow[flow_perc[column_name] <= mol_perc].index)

        if input_stream is not None:

            if rel_perc_direction == "ge":
                index_2 = set(out_flow[flow_perc_rel_to_inp[column_name] >= rel_perc].index)
            else:
                index_2 = set(out_flow[flow_perc_rel_to_inp[column_name] <= rel_perc].index)

            index = pd.Index(index_1 & index_2)
        else:
            index = pd.Index(index_1)

    else:

        flow_max = flow_perc.max(axis=1)

        if mol_perc_direction == "ge":
            index_1 = set(out_flow[flow_max >= mol_perc].index)
        else:
            index_1 = set(out_flow[flow_max <= mol_perc].index)

        if input_stream is not None:
            flow_max_rel_to_inp = flow_perc_rel_to_inp.max(axis=1)

            if rel_perc_direction == "ge":
                index_2 = set(out_flow[flow_max_rel_to_inp >= rel_perc].index)
            else:
                index_2 = set(out_flow[flow_max_rel_to_inp <= rel_perc].index)

            index_3 = set(out_flow[flow_perc.idxmax(axis=1) == flow_perc_rel_to_inp.idxmax(axis=1)].index)
            index = pd.Index(index_1 & index_2 & index_3)
        else:
            index = pd.Index(index_1)

    return index


def update_material_streams(streams: Dict[str, dict]):
    """Add dict data to stream using object_id."""
    outputs = []

    for stream_id, data in streams.items():
        outputs.append(registry.get_element(stream_id)(**data))

    return tuple(outputs)



import os
from sys import platform

from ml4pd._components import components
from ml4pd._registry import registry

# Try adding GraphViz executable to path for windows
if platform == "win32":
    current_paths = os.environ["PATH"].split(os.pathsep)
    for path in [r"C:\Program Files\Graphviz\bin", r"C:\Program Files (x86)\Graphviz\bin"]:
        if path not in current_paths:
            try:
                files = os.listdir(path)
                if "dot.exe" in files:
                    os.environ["PATH"] += os.pathsep + path
                    break
            except FileNotFoundError:
                pass



"""Contains flowsheet object for connecting unit ops."""

import warnings
from typing import Dict, List, Union

import networkx as nx
from pydantic import validate_arguments

try:
    import graphviz as gv
    from graphviz import Source
except ImportError:
    warnings.warn("Couldn't import graphviz. plot_model() not usable.", stacklevel=2)

from ml4pd import registry
from ml4pd.aspen_units.unit import UnitOp
from ml4pd.streams.stream import Stream


class Flowsheet:
    """Conceptual representation of flowsheets in Aspen.

    ```python
    Flowsheet(
        input_streams: List[Stream],
        output_streams: List[Stream],
        object_id: str = None
    )
    ```

    !!! note
        `Flowsheet` can only currently handle tree-like process
        models (one input stream with no loops).

    ## Data Checks
    - If passed graph isn't fully connected, an error will be raised.

    ## Methods

    - `Flowsheet.plot_model()`: returns a graphviz digraph object.
    - `Flowsheet.run(inputs: Dict[str, dict])`: Perform simulation. See [Example](#example).
    - `Flowsheet.clear_data()`: clear both input & output from streams & unit ops.
    - `Flowsheet.get_element(object_id: str)`: returns stream or unit op given its object_id.
    - `Flowsheet._get_networkx(): returns networkx object that can be manually manipulated for plotting/debugging.

    ## Example
    ```python
    from ml4pd import components
    from ml4pd.flowsheet import Flowsheet
    from ml4pd.streams import MaterialStream
    from ml4pd.aspen_unit_ops import Distillation

    components.set_components(["water", "ethanol", "acetone"])
    molecules = {"name_A": ["water", "acetone"], "name_B": ["ethanol", "water"]}
    flowrates = {"flowrate_A": [0.5, 0.3], "flowrate_B": [0.7, 0.7]}

    feed_stream = MaterialStream(object_id="feed")
    dist_column = Distillation(object_id="dist")
    bott_stream, dist_stream = dist_column(feed_stream)

    inputs = {
        "feed": {
            "vapor_fraction": 0,
            "pressure": [3, 4],
            "molecules": molecules,
            "flowrates": flowrates
        },
        "dist": {
            "no_stages": [5, 6],
            "pressure": [3, 4],
            "reflux_ratio": [1, 2],
            "boilup_ratio": [1, 2],
            "feed_stage": [2, 2]
        }
    }

    flowsheet = Flowsheet(input_streams=[feed_stream], output_streams=[bott_stream, dist_stream])
    flowsheet.run(inputs=inputs)

    ```
    """

    unit_no = -1

    @validate_arguments
    def __init__(
        self,
        input_streams: List[Stream],
        output_streams: List[Stream],
        object_id: str = None,
    ):
        self.input_streams = input_streams
        self.output_streams = output_streams
        self.passed = False

        Flowsheet.unit_no += 1
        if object_id is None:
            self.object_id = f"flowsheet_{Flowsheet.unit_no}"
        else:
            self.object_id = id

        self.node_attributes = {
            "distillation": {"shape": "rectangle", "fillcolor": "#ADD8E6", "style": "filled"},
            "flash": {"shape": "triangle", "fillcolor": "#ADD8E6", "style": "filled"},
        }

        # Get networkx graph and add output nodes.
        graphs = [registry.get_element(stream.before).networkx for stream in self.output_streams]
        graph = self._check_connections(graphs)
        self.networkx = self._add_output_nodes(graph)

    def plot_model(self) -> gv.Digraph:
        """Return graphviz object. If in .ipynb, the object is automatically plotted."""

        pydot_graph = nx.nx_pydot.to_pydot(self.networkx.copy())
        pydot_graph.set_rankdir("LR")

        for data_pair in list(self.networkx.nodes(data=True)):
            node = data_pair[0]
            data = data_pair[1]

            if "type" in data:
                node_type = data["type"]
                attributes = self.node_attributes[node_type]
                pydot_graph.get_node(node)[0].set_shape(attributes["shape"])
                pydot_graph.get_node(node)[0].set_fillcolor(attributes["fillcolor"])
                pydot_graph.get_node(node)[0].set_style(attributes["style"])

        return Source(pydot_graph.to_string())

    def _check_connections(self, graphs: List[nx.DiGraph]) -> nx.DiGraph:
        """Check if the graph is fully connected.

        Returns:
            nx.DiGraph: common graph if all checks pass.
        """

        if not all(graph == graphs[0] for graph in graphs):
            raise ValueError("Disconnected graph.")

        graph = graphs[0].copy()

        input_ids = [stream.object_id for stream in self.input_streams]
        edge_labels = [edge[2]["label"] for edge in graph.edges(data=True)]
        if not all(object_id in edge_labels for object_id in input_ids):
            raise ValueError("Disconnected graph.")

        return graph

    def _add_output_nodes(self, graph: Union[nx.DiGraph, gv.Digraph]):
        """_summary_

        Args:
            graph (Union[nx.DiGraph, gv.Digraph]): _description_

        Returns:
            _type_: _description_
        """

        new_graph = graph.copy()

        if isinstance(new_graph, gv.Digraph):
            for stream in self.output_streams:
                new_graph.edge(stream.before, f"{stream.object_id}_output", label=stream.object_id)
        elif isinstance(new_graph, nx.DiGraph):
            for stream in self.output_streams:
                new_graph.add_edge(stream.before, f"{stream.object_id}_output", label=stream.object_id)

        return new_graph

    @validate_arguments
    def run(self, inputs: Dict[str, dict]):
        """Pass data through the process model and perform predictions.

        Args:
            inputs (Dict[str, dict]): a nested dictionary. primary key is the
                object_id of stream/unit op that will receive the inner dict.

        Raises:
            RuntimeError: if flowsheet has been run and the data needs to be cleared.
            ValueError: if therea re multiple input streams.
        """

        if self.passed:
            raise RuntimeError("Please call .clear_data() before attempting a rerun.")
        if len(self.input_streams) != 1:
            raise ValueError("Can only handle 1 input stream at the moment.")

        stream_id = self.input_streams[0].object_id
        _ = self.input_streams[0](**inputs[stream_id])
        first_node = self._get_next_nodes(f"{stream_id}_input")[0]

        self._travel_graph(inputs, first_node)

    def _get_next_nodes(self, node_id: str) -> List[str]:
        """Find successors of node_id that aren't output nodes.

        Args:
            node_id (str): object_id of node to find successors for.

        Returns:
            List[str]: list of object_id of successors of node_id.
        """
        next_nodes = list(self.networkx.successors(node_id))
        next_nodes = [node for node in next_nodes if "output" not in node]
        return next_nodes

    def _travel_graph(self, inputs: Dict[str, dict], node_id: str):
        """Iterative Depth-First Search of a directed graph.

        Args:
            inputs (Dict[str, dict]): a nested dictionary. primary key is the
                object_id of stream/unit op that will receive the inner dict.
            node_id (str): object_id of current node.
        """
        current_node = registry.get_element(node_id)
        if current_node.data is None:
            stream = registry.get_element(current_node.before[0])
            _ = current_node(stream, **inputs[current_node.object_id])

            next_nodes = self._get_next_nodes(current_node.object_id)
            for node in next_nodes:
                self._travel_graph(inputs, node)

    def _get_initial_streams(self, input_streams: List[Stream]) -> List[str]:
        """Get streams that have to be fed first, as opposed to those that come in mid-prrocess."""

        adj = self.networkx.adj
        initial_streams = []
        for stream in input_streams:
            stream_id = stream.object_id
            aspen_unit = list(dict(adj[stream_id]).keys())[0]
            if len(list(self.networkx.predecessors[aspen_unit])) == 1:
                initial_streams.append(stream_id)

        if len(initial_streams) == 0:
            raise ValueError("No initial streams found.")

        return initial_streams

    def _get_networkx(self) -> nx.DiGraph:
        """Return networkx object primarily for debugging."""
        return self.networkx

    def clear_data(self):
        """Loop through the graph to clear streams & unit ops data."""
        graph = self.networkx.copy()
        edges = list(graph.edges(data=True))
        nodes = dict(graph.nodes(data=True))

        for edge in edges:
            stream = registry.get_element(edge[2]["label"])
            stream.clear_data()
        for node in nodes:
            try:
                column = registry.get_element(node)
                column.clear_data()
            except KeyError:
                pass

    @validate_arguments
    def get_element(self, object_id: str) -> Union[Stream, UnitOp]:
        """Return object of object_id, if it's in the graph.

        Args:
            object_id (str): object_id of object to search for.

        Returns:
            Union[Stream, UnitOp]: stream of unit op with object_id.
        """

        edge_labels = [edge[2]["label"] for edge in self.networkx.edges(data=True)]

        if self.networkx.has_node(object_id):
            element = registry.get_element(object_id)
        elif object_id in edge_labels:
            element = registry.get_element(object_id)
        else:
            raise ValueError("object_id not found in graph.")

        return element



"""Predict the aspen output of a flash column."""


import importlib.resources
import pickle
from dataclasses import field
from typing import ClassVar, Dict, List, Literal, Tuple, Union

import numpy as np
from ml4pd.aspen_units.unit import UnitOp
from ml4pd.streams import MaterialStream
from ml4pd.aspen_units.utils import get_model, relu
from ml4pd.streams.utils import update_material_streams
from ml4pd.utils import timer

try:
    from ml4pd_models import flash_models
except ModuleNotFoundError:
    print("Couldn't find ml4pd_models. Check https://github.com/NREL/ml4pd_models.")
from pydantic import validate_arguments
from pydantic.dataclasses import dataclass


@dataclass(eq=False)
class Flash(UnitOp):
    """
    Emulates Aspen's flash column. It gathers data from feed stream,
    and combine additional unit-specific input to make predictions on
    duties and vapor & liquid streams.

    ```python
    Flash(
        pressure: Union[float, List[float]] = None
        temperature: Union[float, List[float]] = None
        duty: Union[float, List[float]] = None
        vapor_fraction: Union[float, List[float]] = None
        valid_phases: Literal["vl", "vll", "vl-fw", "vl-dw"] = None
        pres_units: str = "atm"
        temp_units: Literal["degC", "degF", "degK", "degR"] = "degC"
        duty_units: str = "kJ/hr"
        property_method: Literal["nrtl"] = "nrtl"
        verbose: bool = False
        check_data: bool = True
        object_id: str = None
    )(feed_stream: MaterialStream, **kwargs)
    ```

    ## Data Checks & Manipulations

    - 2 out of `pressure`, `temperature`, `duty`, `vapor_fraction` must be specified.
    - Will raise an error if pressure drop results in `pressure < 0`.
    - Input must obey type hints (enforced by Pydantic).

    ## Support for Batch Predictions
    `Distillation` currently has limtied support for broadcasting when

    - The incoming feed stream has 1 row of data while current unit data has multiple.
    - The incoming feed stream has multiple rows of data while the current unit data has 1.

    The best way to to take advantage of batch predictions is to prepare a dataframe of
    input parameteras and input individual columns.

    ## Useful Attributes

    - `object_id`: unique id of object. Can be set during initialization, and will be checked
    against a global registry to prevent duplicates.
    - `before`: list of feed streams.
    - `after`: list of output streams.
    - `data`: a pandas df containing all data used for ML.
    - `status`: numpy array specifying which row of ML df should be trusted. Similar to how
    Aspen indicates Errors.

    ## Example
    ```python
    from ml4pd import components
    from ml4pd.streams import MaterialStream
    from ml4pd.aspen_unit_ops import Flash

    components.set_components(["water", "ethanol", "acetone"])
    molecules = {"name_A": ["water", "acetone"], "name_B": ["ethanol", "water"]}
    flowrates = {"flowrate_A": [0.5, 0.3], "flowrate_B": [0.7, 0.7]}

    feed = MaterialStream(pressure=3, vapor_fraction=0.4)
    flash = Flash(pressure=3, duty=0)

    feed = feed(molecules=molecules, flowrates=flowrates)
    vapor, liquid = flash(feed)
    ```

    """

    unit_no: ClassVar[int] = -1

    # ------ DO NOT CHANGE THESE VARIABLES. WILL MESS UP ML. ------ #
    pressure: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    temperature: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    duty: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    vapor_fraction: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    valid_phases: Literal["vl", "vll", "vl-fw", "vl-dw"] = field(default="vl", repr=False, metadata={"grp": "str"})
    # ------ DO NOT CHANGE THESE VARIABLES. WILL MESS UP ML. ------ #

    pres_units: str = "atm"
    temp_units: Literal["degC", "degF", "degK", "degR"] = "degC"
    duty_units: str = "kJ/hr"
    property_method: Literal["nrtl"] = "nrtl"

    def __post_init__(self):

        # Add unique id to column and log it in registry.
        Flash.unit_no += 1
        if self.object_id is None:
            self.object_id = f"F{Flash.unit_no}"

        self.after: Dict[str, str] = {"vapor": None, "liquid": None}

        super().__post_init__()

    @validate_arguments
    def __call__(self, feed_stream: MaterialStream, **kwargs) -> Tuple[MaterialStream, MaterialStream]:

        for key, value in kwargs.items():
            if key not in self.__dataclass_fields__:
                raise AttributeError(f"{key} not recognized.")
            setattr(self, key, value)

        if feed_stream.data is not None:

            if self.check_data:
                with timer(verbose=self.verbose, operation="data check", unit=self.object_id) as _:
                    self._check_redundancy()
                    self._check_pressure(feed_stream)
                    self._check_units()
            with timer(verbose=self.verbose, operation="data prep", unit=self.object_id) as _:
                self._adjust_pressure(feed_stream)
                self.unit_data = self._format_unit_data()
                self.data = self._combine_unit_and_stream_data(feed_stream)

            with timer(verbose=self.verbose, operation="ML", unit=self.object_id) as _:
                vapor_stream, liquid_stream = self._predict(feed_stream)

        else:
            vapor_stream = MaterialStream(stream_type="vapor", before=self.object_id)
            liquid_stream = MaterialStream(stream_type="liquid", before=self.object_id)

        self._add_to_graph(feed_stream, "flash")
        self._update_connections(input_stream=feed_stream, output_streams={"vapor": vapor_stream, "liquid": liquid_stream})

        return vapor_stream, liquid_stream

    def _predict(self, feed_stream: MaterialStream) -> Tuple[MaterialStream, MaterialStream]:

        model_fname = get_model(module=flash_models, pattern=f"flash_{len(feed_stream._suffixes)}_")

        with importlib.resources.path(flash_models, model_fname) as model_path:
            with open(model_path, "rb") as model_file:
                stat_model = pickle.load(model_file)
                flow_model = pickle.load(model_file)
                temp_model = pickle.load(model_file)

        # Prepare data
        if self.fillna:
            x = self.data.fillna(self.na_value).select_dtypes(exclude="object")
        else:
            x = self.data.select_dtypes(exclude="object")

        # Get predictions
        stat = stat_model.predict(x)
        flow = flow_model.predict(x)
        temp = temp_model.predict(x)

        # Get flowrates
        liquid_flow_perc = relu(flow, max_value=1.0)
        liquid_flow = feed_stream.flow * liquid_flow_perc
        vapor_flow = feed_stream.flow - liquid_flow

        # Turn flowrates into dictionaries with correct keys.
        col_dict = dict(zip(feed_stream._flow_columns, [f"flowrate_{suffix}" for suffix in feed_stream._suffixes]))
        liquid_flow = liquid_flow.rename(columns=col_dict).to_dict("list")
        vapor_flow = vapor_flow.rename(columns=col_dict).to_dict("list")

        # set status for column
        if feed_stream.status is not None:
            self.status = stat * feed_stream.status
        else:
            self.status = stat

        # Prepare streams
        vapor_data = {
            "vapor_fraction": 1,
            "temperature": temp.mean(axis=1).tolist(),
            "molecules": feed_stream.data[feed_stream._name_columns].to_dict("list"),
            "flowrates": vapor_flow,
        }

        liquid_data = {
            "vapor_fraction": 0,
            "temperature": temp.mean(axis=1).tolist(),
            "molecules": feed_stream.data[feed_stream._name_columns].to_dict("list"),
            "flowrates": liquid_flow,
        }

        if None not in self.after.values():
            vapor_stream, liquid_stream = update_material_streams(streams={self.after["vapor"]: vapor_data, self.after["liquid"]: liquid_data})

        else:
            vapor_stream = MaterialStream(stream_type="vapor")(**vapor_data, before=self.object_id)
            liquid_stream = MaterialStream(stream_type="liquid")(**liquid_data, before=self.object_id)

        vapor_stream.status = self.status
        liquid_stream.status = self.status

        return vapor_stream, liquid_stream

    def _adjust_pressure(self, feed_stream: MaterialStream):
        """If flash pressure is 0, set it equal to feed pressure."""

        if isinstance(self.pressure, list):
            selector = (np.array(self.pressure) == 0) * 1
            new_pressure = self.pressure + (np.array(feed_stream.pressure) * selector)
            self.pressure = new_pressure.tolist()
        elif self.pressure == 0:
            self.pressure = feed_stream.pressure

    def _check_redundancy(self):
        """Only 2 out of 4 numerical variables can be set."""

        input_sum = sum([element is not None for element in [self.pressure, self.temperature, self.duty, self.vapor_fraction]])
        if input_sum != 2:
            raise ValueError("Specify exactly 2 out of pressure, temperature, duty, and vapor fraction.")

    def _check_pressure(self, feed_stream: MaterialStream):
        """Check that pressure drops don't lead to negative pressures."""

        if self.pressure is not None and feed_stream.pressure is not None:
            selector = (np.array(self.pressure) < 0) * 1
            test_flash_pressure = selector * self.pressure
            test_feed_pressure = selector * feed_stream.pressure
            test_pressure = test_feed_pressure + test_flash_pressure
            if (test_pressure < 0).sum() != 0:
                raise ValueError("Specified pressure drop would result in a pressure less than 0.")



"""Simulate aspen material streams."""

import warnings
from dataclasses import field
from functools import partial
from typing import Dict, List, Literal, Union

import numpy as np
import pandas as pd
from pint import UnitRegistry
from pydantic import validate_arguments
from pydantic.dataclasses import dataclass

from ml4pd import components
from ml4pd.streams.stream import Stream
from ml4pd.utils import timer

ureg = UnitRegistry()

__all__ = ["MaterialStream"]


@dataclass(eq=False)
class MaterialStream(Stream):
    """
    Emulates an Aspen material stream. Its primary purpose is to get **molecular**,
    **flowrate** and **state data** from user and perform data checks before passing this
    information along to other unit ops.

    ```python

    MaterialStream(
        temperature: Union[float, List[float]] = None
        vapor_fraction: Union[float, List[float]] = None
        pressure: Union[float, List[float]] = None
        stream_type: Literal["feed", "bott", "dist", "vapor", "liquid"] = "feed"
        pressure_units: str = "atm"
        flowrates_units: str = "kmol/hr"
        temperature_units: Literal["degC", "degF", "degK", "degR"] = "degC"
        check_data: bool = True
        verbose: bool = False
        object_id: str = None
        before: str = None
        after: str = None
    )(
        molecules: Dict[str, Union[str, List[str]]],
        flowrates: Dict[str, Union[float, List[float]]],
        **kwargs,
    )

    ```

    ## Data Checks & Manipulations

    - 2 out of `temperature`, `vapor_fraction`, and `temperature` must be specified.
    - User can input the units of the input data via `flowrate_units`, `temperature_units`,
    and `pressure_units`. `MaterialStream` will then convert the quantities to the
    default units so that they agree with training data.
    - For `flowrates`, errors will be raised if there are `NaN`, or negative values. Also,
    keys to this dictionary must be of `f"flowrate_{suffix}"` format.
    - For  `molecules,` errors will be raised if there are NaN or the specified
    molecules are not found in `components`. Also, keys to this dictionary must be of
    `f"name_{suffix}"` format.
    - For each row, sort `molecules` & `flowrates` so that the lightest molecules are to the left.
    - Input must obey type hints (which is enforced by Pydantic).

    ## Support for Batch Predictions
    `MaterialStream` currently has limited support for broadcasting when

    - there is 1 set of `molecules` and multiple sets of `flowrates`.
    - there are n sets of `molecules` and `flowrates` and 1 set of state variables.

    The best way to to take advantage of batch predictions is to prepare a dataframe of
    input parameters and input individual columns.

    ## Useful Attributes
    - `object_id`: unique id of object. Automatically generated but can be set during
    initialization, and will be checked against a global registry to prevent duplicates.
    - `before`: id of the unit that produced this stream.
    - `after`: id of the unit that this stream feeds into.
    - `data`: a pandas df containing all data used for ML.
    - `temperature`: predicted if instance is produced by a unit operation.
    - `flow`: predicted if instance is produced by a unit operation.

    ## Vapor Fraction/Temperature/Pressure
    - To reduce the number of ML models needed, `MaterialStream` created by ML will only predict
    temperature & vapor fraction since only 2 of 3 variables are needed to fully specify the stream.
    - User-created `MaterialStream` currently only suspports vapor fraction & temperature or
    vapor fraction & pressure. The range of these variables depend on the unit op the stream feeds to.

    ## Example
    !!! note
        `feed(molecules=molecules, flowrates=flowrates, **kwargs)` must be used to create/update `data`.

    ```python
    from ml4pd import components
    from ml4pd.streams import MaterialStream
    from ml4pd.aspen_unit_ops import Distillation

    components.set_components(["water", "ethanol", "acetone"])
    molecules = {"name_A": ["water", "acetone"], "name_B": ["ethanol", "water"]}
    flowrates = {"flowrate_A": [0.5, 0.3], "flowrate_B": [0.7, 0.7]}

    feed = MaterialStream(pressure=3, vapor_fraction=0)(molecules=molecules, flowrates=flowrates)
    ```

    """

    # ------ DO NOT CHANGE THE ORDER OF THESE VARIABLES. WILL MESS UP ML. ------ #
    temperature: Union[float, List[float]] = field(default=None, repr=False)
    vapor_fraction: Union[float, List[float]] = field(default=None, repr=False)
    pressure: Union[float, List[float]] = field(default=None, repr=None)
    # ------ DO NOT CHANGE THE ORDER OF THESE VARIABLES. WILL MESS UP ML. ------ #

    stream_type: Literal["feed", "bott", "dist", "liquid", "vapor"] = "feed"
    pressure_units: str = "atm"
    flowrates_units: str = "kmol/hr"
    temperature_units: Literal["degC", "degF", "degK", "degR"] = "degC"

    def __post_init__(self):
        """
        Ensure every newly instantiated stream has a unique identifier &
        add instance to a registry to later be used to graph modelling.
        """

        self.flow: pd.DataFrame = None
        self.flow_norm: pd.DataFrame = None
        self.flow_sum: pd.Series = None
        self._flow_columns: List[str] = []
        self._state_columns: List[str] = []
        self._suffixes: List[str] = None
        self._name_columns: List[str] = None
        self._mw_idx: np.ndarray = None

        MaterialStream.unit_no += 1
        if self.object_id is None:
            self.object_id = f"{self.stream_type}{MaterialStream.unit_no}"

        super().__post_init__()

    @validate_arguments
    def __call__(
        self,
        molecules: Dict[str, Union[str, List[str]]],
        flowrates: Dict[str, Union[float, List[float]]],
        **kwargs,
    ):
        """
        Get molecules & flowrates data then assemble df after multiple checks.

        Args:
            molecules (Dict[str, Union[str, List[str]]]): dictionary of
                molecules w/ 'name_{suffix}' format for columns.
            flowrates (Dict[str, Union[float, List[float]]]): dictionary of
                corresponding flowrates w/ 'flowrate_{suffix}' format for columns.

        Raises:
            AttributeError: if keys in kwargs aren't in __init__.

        Returns:
            MaterialStream: current instance to be fed into columns.
        """

        for key, value in kwargs.items():
            if key not in self.__dataclass_fields__:
                raise AttributeError(f"{key} not recognized.")
            setattr(self, key, value)

        try:
            self.data = pd.DataFrame(molecules)
        except ValueError:
            self.data = pd.DataFrame({key: [value] for key, value in molecules.items()})

        self._name_columns = self.data.columns.to_list()
        self._suffixes = [col.split("_")[-1] for col in self.data.columns]

        try:
            self.flow = pd.DataFrame(flowrates)
        except ValueError:
            self.flow = pd.DataFrame({key: [value] for key, value in flowrates.items()})

        if self.check_data:
            with timer(verbose=self.verbose, operation="data check", unit=self.object_id) as _:
                self._check_state_variables()
                self._check_prefix()
                self._check_suffix()
                self._check_flowrates()
                self._check_molecules()
                self._convert_units()

        with timer(verbose=self.verbose, operation="data prep", unit=self.object_id) as _:
            self.flow_sum = self.flow.sum(axis=1)
            self.flow_norm = self.flow.div(self.flow_sum, axis=0)
            self._broadcast_molecules_and_flowrates()
            self._sort_molecules_by_weight()
            self._add_mol_data()
            self._add_flowrates_and_state_variables()
            self._add_comp_no()

        return self

    def clear_data(self):
        """Clear user-defined data. Most useful for flowsheets."""

        # Delete state variables
        delattr(self, "temperature")
        delattr(self, "vapor_fraction")
        delattr(self, "pressure")

        # Delete data and flow
        delattr(self, "data")
        delattr(self, "flow")

        # Delete private variables
        delattr(self, "_suffixes")
        delattr(self, "_flow_columns")
        delattr(self, "_state_columns")
        delattr(self, "_mw_idx")

    def _sort_molecules_by_weight(self):
        """
        Sort molecular names & flowrates by molecular weights.
        It will also set the '_mw_idx' attribute to later be used on training
        data's output.
        """

        data_copy = self.data.copy(deep=True)

        # Get molecular weights to sort
        molecular_weight = components.data[["name", "MolWt"]]
        molecular_weight_columns = []
        name_columns = []
        for suffix in self._suffixes:
            molecular_weight_columns.append(f"MolWt_{suffix}")
            name_columns.append(f"name_{suffix}")
            mol_data = molecular_weight.rename(columns={"name": f"name_{suffix}", "MolWt": f"MolWt_{suffix}"})
            data_copy = pd.merge(data_copy, mol_data, how="left")

        # Sort molecules and flowrates by sorted indices of molecular weights
        molecular_weight_idx = data_copy[molecular_weight_columns].to_numpy().argsort()
        data_copy = np.take_along_axis(data_copy[name_columns].to_numpy(), molecular_weight_idx, axis=1)
        flowrates = np.take_along_axis(self.flow.to_numpy(), molecular_weight_idx, axis=1)
        flowrates_norm = np.take_along_axis(self.flow_norm.to_numpy(), molecular_weight_idx, axis=1)

        # Replace input data with sorted data
        self.data = pd.DataFrame(data_copy, columns=self.data.columns)
        self.flow = pd.DataFrame(flowrates, columns=self.flow.columns)
        self.flow_norm = pd.DataFrame(flowrates_norm, columns=self.flow_norm.columns)

        # Add idx to instance for other uses like sorting y-values during training
        self._mw_idx = molecular_weight_idx

    def _convert_units(self):
        """Convert units of pressure, temperature, and flowrates into default ones."""

        if self.pressure_units != "atm" and self.pressure is not None:
            self.pressure = (self.pressure * ureg(self.pressure_units)).to("atm").magnitude
            self.pressure_units = "atm"

        if self.temperature_units != "degC" and self.temperature is not None:
            original_quant = ureg.Quantity(self.temperature, self.temperature_units)
            self.temperature = original_quant.to("degC").magnitude
            self.temperature_units = "C"

        if self.flowrates_units != "kmol/hr" and self.flow is not None:
            original_quant = self.flow.to_numpy() * ureg(self.flowrates_units)
            default_quant = original_quant.to("kmol/hr").magnitude
            self.flow = pd.DataFrame(default_quant, columns=self.flow.columns)
            self.flowrates_units = "kmol/hr"

    def _broadcast_molecules_and_flowrates(self):
        """Attempt to match the shapes of molecules & flowrates.

        Raises:
            ValueError: if len(data) != len(flow) and neither is 1.
        """

        if len(self.data) == 1 and len(self.flow_norm) != 1:
            warnings.warn("will broadcast molecules to flowrates.", stacklevel=2)
            self.data = pd.concat([self.data] * len(self.flow_norm), ignore_index=True)
        elif len(self.data) != 1 and len(self.flow_norm) == 1:
            warnings.warn("will broadcast flowrates to molecules.", stacklevel=2)
            self.flow_norm = pd.concat([self.flow_norm] * len(self.data), ignore_index=True)
        elif len(self.flow_norm) != len(self.flow_norm):
            raise ValueError("Molecules & Flowrates shape don't match.")

    def _add_flowrates_and_state_variables(self):
        """Add flowrates & state variables to data, with some support for broadcasting.

        Raises:
            ValueError: if data & flowrates are not of the right shape.
        """

        for suffix in self._suffixes:
            self.data[f"{self.object_id}_flowrate_{suffix}"] = self.flow_norm[f"flowrate_{suffix}"]
            self._flow_columns.append(f"{self.object_id}_flowrate_{suffix}")

        for col in ["temperature", "vapor_fraction", "pressure"]:
            self.data[f"{self.object_id}_{col}"] = getattr(self, col)
            self._state_columns.append(f"{self.object_id}_{col}")

    def _add_comp_no(self):
        """Add number of molecular types in a stream."""

        self.data[f"{self.object_id}_comp_no"] = (self.data[self._flow_columns] != 0.0).sum(axis=1)

    def _add_mol_data(self):
        """Get molecular data from components object and add them self.data."""

        def add_suffix(col, suf):
            return col + suf

        for suffix in self._suffixes:
            mol_data = components.data.rename(mapper=partial(add_suffix, suf=f"_{suffix}"), axis="columns")
            self.data = pd.merge(self.data, mol_data, how="left")

    def _check_prefix(self):
        """Check prefixes of molecules & flowrates dictionaries.

        Raises:
            ValueError: if prefix of molecules isn't 'name.'
            ValueError: if prefix of flowrates isn't 'flowrate.'
        """

        molecules_prefix = [col.split("_")[0] for col in self.data.columns]
        flowrates_prefix = [col.split("_")[0] for col in self.flow.columns]

        if (len(set(molecules_prefix)) != 1) or ("name" not in set(molecules_prefix)):
            raise ValueError("Only 1 allowed prefix for molecules columns: name.")
        if (len(set(flowrates_prefix)) != 1) or ("flowrate" not in set(flowrates_prefix)):
            raise ValueError("Only 1 allowed prefix for flowrates columns: flowrate.")

    def _check_suffix(self):
        """Check suffixes of molecules  flowrates dictionaries.

        Raises:
            ValueError: if suffixes in molecules aren't unique.
            ValueError: if suffixes in flowrates aren't unique.
            ValueError: if suffixes in flowrates & molecules don't agree.
        """

        molecules_suffix = [col.split("_")[-1] for col in self.data.columns]
        flowrates_suffix = [col.split("_")[-1] for col in self.flow.columns]

        if len(set(molecules_suffix)) != len(molecules_suffix):
            raise ValueError("Suffixes in molecules must be unique.")
        if len(set(flowrates_suffix)) != len(flowrates_suffix):
            raise ValueError("Suffixes in flowrates must be unique.")
        if set(flowrates_suffix) != set(molecules_suffix):
            raise ValueError("Suffixes in molecules & flowrates must agree.")

    def _check_state_variables(self):
        """Check that state variables make sense.

        Raises:
            ValueError: If more or less than 2 of them are specified.
        """

        state_variables = [self.temperature, self.vapor_fraction, self.pressure]
        if sum([var is not None for var in state_variables]) != 2:
            raise ValueError("Expected exactly 2 out of state variables.")

    def _check_flowrates(self):
        """Check that flowrates make sense and normalize them if specified by user.

        Raises:
            ValueError: if there are NaN in flowrates.
            ValueError: if there are flowrates smaller than 0.
        """

        if self.flow.isna().sum().sum() != 0:
            raise ValueError("Found NaN in flowrates.")

        if (self.flow < 0).sum().sum() > 0:
            raise ValueError("flowrates cannot be smaller than 0.")

    def _check_molecules(self):
        """Check that molecules make sense.

        Raises:
            ValueError: if there are NaN in molecules.
            ValueError: if molecules aren't pre-specified in components.
        """

        if self.data.isna().sum().sum() != 0:
            raise ValueError("Found NaN in molecules.")

        unique_molecules = set(self.data.to_numpy().flatten())
        if len(unique_molecules - set(components.data["name"])) != 0:
            print(set(components.data["name"]))
            raise ValueError(f"{unique_molecules -  set(components.data['name'])} not in components.")



"""
Module gives a components object that a user can use to specify
the types of molecules to consider in the system.
"""

import importlib.resources
import json
import warnings
from typing import List

import pandas as pd
import thermo
from pydantic import BaseModel, StrictStr, validate_arguments
from rdkit.Chem import Descriptors, MolFromSmiles

from ml4pd import data

with importlib.resources.path(data, "rdkit_features.txt") as rdkit_features_path:
    with open(rdkit_features_path, "r") as rdkit_features_file:
        rdkit_features = [line.strip() for line in rdkit_features_file.readlines()]

with importlib.resources.path(data, "thermo_dict.json") as thermo_dict_path:
    with open(thermo_dict_path, "r") as thermo_dict_file:
        thermo_dict = json.load(thermo_dict_file)

__all__ = ["components"]


class Components(BaseModel):
    """
    Object to specify molecules in the system and pass that information along to streams and unit ops.

    ## Useful Attributes:

    - `data`: a df where each row is a molecule and each column is a feature of that molecule.
    - `__repr__`: table where each row is a specified molecule, and each column is a type of identifier.
        Useful for checking that `components` got the right molecules.

    ## Example
    ```python
    from ml4pd import components
    components.set_components(["acetone", "water", "ethanol"])
    ```
    """

    class Config:
        arbitrary_types_allowed = True

    data: pd.DataFrame = None
    fill_na: bool = True

    @validate_arguments
    def set_components(self, molecules: List[StrictStr]):
        """
        Specify molecules to use and get data for. Streams & unit ops will get
        data from here & check their input data against the specified molecules.

        Args:
            molecules (pd.DataFrame, dict, list): if pd.DataFrame/dict, the
            key/column must be 'name.' The molecules must have 'iupac-like'
            names. Don't input formula/cas/smiles or other representations.
            For example, to specify 2-butanone, butan-2-one also works.
        """

        if len(molecules) != len(set(molecules)):
            warnings.warn("Dropping duplicates found in molecules.", stacklevel=2)
            molecules = list(set(molecules))

        if len(molecules) < 2:
            warnings.warn("There are less than two molecules in components.", stacklevel=2)

        self.data = pd.DataFrame({"name": molecules})
        identifiers = self._get_identifiers()
        self.data = self.data.merge(identifiers, left_on="name", right_on="name")
        self.data = self.data.apply(self._get_rdkit_data, axis=1)
        self.data = self._get_thermo_data(self.data).rename(columns=thermo_dict)
        if self.fill_na:
            nan_cols = self.data.columns[self.data.isna().any()].to_list()
            for col in nan_cols:
                self.data[col] = self.data[col].fillna(self.data[col].median())

    @validate_arguments
    def add_components(self, new_molecules: List[StrictStr]):
        """
        Add more molecules to the existing list of molecules. Use ony after set_components().
        May be ueful when process produces molecules that are not accounted for initially.
        """

        new_molecules = list(set(new_molecules + self.data["name"].to_list()))
        self.set_components(new_molecules)

    def _get_identifiers(self) -> pd.DataFrame:
        """
        Looks up different identifiers for the input molecules.

        Returns:
            identifiers (pd.DataFrame): with columns 'name', 'cas', 'smiles', 'iupac.'
        """
        identifiers = thermo.datasheet.tabulate_constants(self.data["name"], full=True)[
            ["smiles", "IUPAC name", "CAS"]
        ].reset_index(drop=True)
        identifiers["name"] = self.data["name"]
        identifiers = identifiers.rename(columns={"IUPAC name": "iupac", "CAS": "cas"})

        return identifiers

    @staticmethod
    def _get_rdkit_data(row: pd.Series) -> pd.Series:
        """
        For each row, look up rdkit data specified by the list rdkit_features.

        Args:
            row (pd.Series): must contain a 'smiles' column.

        Returns:
            (pd.Series): molecule + rdkit features.
        """
        smiles = row["smiles"]
        molecule = MolFromSmiles(smiles)

        if molecule:
            fns = [(x, y) for x, y in Descriptors.descList if x in rdkit_features]
            res = {}
            for feature, function in fns:
                res[feature] = function(molecule)
        else:
            res = dict(zip(rdkit_features, [None] * (len(rdkit_features))))
            res["smiles"] = smiles

        return pd.Series(dict(**row, **res))

    @staticmethod
    def _get_thermo_data(df: pd.DataFrame) -> pd.DataFrame:
        """
        Get thermodynamic data from the thermo package for each molecule in df.

        Args:
            df (pd.DataFrame): must contain a "iupac" column.

        Returns:
            df (pd.DataFrame): original df with added thermo columns.
        """
        df_thermo = (
            thermo.datasheet.tabulate_constants(df["name"], full=True)[list(thermo_dict.keys()) + ["IUPAC name"]]
            .reset_index(drop=True)
            .rename(columns={"IUPAC name": "iupac"})
        )

        df = df.merge(df_thermo, left_on="iupac", right_on="iupac").copy(deep=True)

        return df

    def __repr__(self):
        """Returns a prettified string of the identifiers dataframe."""

        return self.data[["name", "iupac", "smiles", "cas"]].to_string()

components = Components()



"""Helper functions for unit ops."""

import importlib.resources
from types import ModuleType
from typing import Union
import numpy as np


def relu(array: np.ndarray, max_value: Union[float, int]) -> np.ndarray:
    """numpy version of tf.ReLU(max_value)

    Args:
        array (np.ndarray): array to apply relu to.
        max_value (Union[float, int]): max value of resulting array.

    Returns:
        np.ndarray: array with same dimentions at array but values are between (0, max_value).
    """
    array = np.maximum(array, np.full(shape=array.shape, fill_value=0))
    array = np.minimum(array, np.full(shape=array.shape, fill_value=max_value))
    return array


def get_model(module: ModuleType, pattern: str, ftype: str = "pkl") -> str:
    """Get ml model from a module.

    Args:
        module (ModuleType): module that contains ml models.
        pattern (str): pick a particular model.
        ftype (str, optional): suffix of model fname. Default is "pkl".

    Returns:
        str: fname of model.
    """

    model_fnames = [fname for fname in list(importlib.resources.contents(module)) if ftype in fname]
    model_fname = [fname for fname in model_fnames if pattern in fname][0]

    return model_fname



"""Predict the aspen output of a distillation column."""

import importlib.resources
import pickle
import warnings
from dataclasses import field
from typing import Dict, List, Literal, Tuple, Union

import numpy as np
from ml4pd.aspen_units.unit import UnitOp
from ml4pd.streams import MaterialStream
from ml4pd.aspen_units.utils import get_model, relu
from ml4pd.streams.utils import update_material_streams
from ml4pd.utils import timer

try:
    from ml4pd_models import distillation_models
except ModuleNotFoundError:
    print("Couldn't find ml4pd_models. Check https://github.com/NREL/ml4pd_models.")

from pydantic import validate_arguments
from pydantic.dataclasses import dataclass


@dataclass(eq=False)
class Distillation(UnitOp):
    """
    Emulates Aspen's RadFrac column. It gathers data from feed stream,
    and combine additional unit-specific input to make predictions on
    duties and bottom & distillate streams.

    ```python
    Distillation(
        no_stages: Union[int, List[int]] None
        pressure: Union[float, List[float]] = None
        dist_rate: Union[float, List[float]] = None
        reflux_ratio: Union[float, List[float]] = None
        bott_rate: Union[float, List[float]] = None
        reflux_rate: Union[float, List[float]] = None
        boilup_rate: Union[float, List[float]] = None
        boilup_ratio: Union[float, List[float]] = None
        dist_to_feed_ratio: Union[float, List[float]] = None
        bott_to_feed_ratio: Union[float, List[float]] = None
        condensor_duty: Union[float, List[float]] = None
        reboiler_duty: Union[float, List[float]] = None
        free_water_reflux_ratio: Union[float, List[float]] = None
        feed_stage: Union[int, List[int]] = None
        calc_type: Literal["equil", "rate-based"] = None
        condensor_type: Literal["total", "partial-v", "partial-vl", "none"] = None
        reboiler_type: Literal["kettle", "thermosiphon", "none"] = None
        valid_phases: Literal["vl", "vll", "vl-fwc", "vl-fwns", "vl-dwc", "vl-dwas"] = None
        convergence: Literal["std", "petro/wb", "snil", "azeo", "cryo", "custom"] = None
        feed_stage_convention: Literal["above", "on", "vapor", "liquid"] = None
        rate_units: str = "kmol/hr"
        duty_units: str = "kJ/hr"
        pres_units: str = "atm"
        property_method: Literal["nrtl"] = "nrtl"
        verbose: bool = False
        check_data: bool = True
        object_id: str = None
    )(feed_stream: MaterialStream, **kwargs)
    ```

    ## Data Checks & Manipulations

    - Will raise an error if `no_stages` and `pressure` aren't specified.
    - If `feed_stage` isn't specified, will default to `round(no_stages/2)`.
    Also, if feed_stage > no_stages + 1, it wll trigger an error.
    - User can input the units of the input data via `rate_units`, `duty_units` and `pres_units`.
    The corresponding quantities will be converted to default units.
    - 2 out of operating configurations (`dist_rate`, `reflux_rate`, etc.) must be specified.
    - Input must obey type hints (enforced by Pydantic).

    ## Support for Batch Predictions
    `Distillation` currently has limtied support for broadcasting when

    - The incoming feed stream has 1 row of data while current unit data has multiple.
    - The incoming feed stream has multiple rows of data while the current unit data has 1.

    The best way to to take advantage of batch predictions is to prepare a dataframe of
    input parameteras and input individual columns.

    ## Useful Attributes

    - `object_id`: unique id of object. Can be set during initialization, and will be checked
    against a global registry to prevent duplicates.
    - `before`: list of feed streams.
    - `after`: list of output streams.
    - `data`: a pandas df containing all data used for ML.
    - `condensor_duty`: predicted given input.
    - `reboiler_duty`: predicted given input.
    - `status`: numpy array specifying which row of ML df should be trusted. Similar to how
    Aspen indicates Errors.

    ## Example
    ```python
    from ml4pd import components
    from ml4pd.streams import MaterialStream
    from ml4pd.aspen_unit_ops import Distillation

    components.set_components(["water", "ethanol", "acetone"])
    molecules = {"name_A": ["water", "acetone"], "name_B": ["ethanol", "water"]}
    flowrates = {"flowrate_A": [0.5, 0.3], "flowrate_B": [0.7, 0.7]}

    feed = MaterialStream(pressure=3, vapor_fraction=0)
    dist_col = Distillation(no_stages=40, feed_stage=10, boilup_ratio=0.5, reflux_ratio=0.5)

    feed = feed(molecules=molecules, flowrates=flowrates)
    bott, dist = dist_col(feed, pressure=3)
    ```

    """

    # ------ DO NOT CHANGE THESE VARIABLES. WILL MESS UP ML. ------ #
    no_stages: Union[int, List[int]] = field(default=None, repr=False, metadata={"grp": "num"})
    pressure: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    dist_rate: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    reflux_ratio: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    bott_rate: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    reflux_rate: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    boilup_rate: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    boilup_ratio: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    dist_to_feed_ratio: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    bott_to_feed_ratio: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    condensor_duty: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    reboiler_duty: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    free_water_reflux_ratio: Union[float, List[float]] = field(default=None, repr=False, metadata={"grp": "num"})
    feed_stage: Union[int, List[int]] = field(default=None, repr=False, metadata={"grp": "num"})
    calc_type: Literal["equil", "rate-based"] = field(default="equil", metadata={"grp": "str"})
    condensor_type: Literal["total", "partial-v", "partial-vl", "none"] = field(default="total", metadata={"grp": "str"})
    reboiler_type: Literal["kettle", "thermosiphon", "none"] = field(default="kettle", metadata={"grp": "str"})
    valid_phases: Literal["vl", "vll", "vl-fwc", "vl-fwns", "vl-dwc", "vl-dwas"] = field(default="vl", metadata={"grp": "str"})
    convergence: Literal["std", "petro/wb", "snil", "azeo", "cryo", "custom"] = field(default="std", metadata={"grp": "str"})
    feed_stage_convention: Literal["above", "on", "vapor", "liquid"] = field(default="above", metadata={"grp": "str"})
    # ------ DO NOT CHANGE THESE VARIABLES. WILL MESS UP ML. ------ #

    rate_units: str = "kmol/hr"
    duty_units: str = "kJ/hr"
    pres_units: str = "atm"
    property_method: Literal["nrtl"] = "nrtl"

    def __post_init__(self):

        # Add unique id to column and log it in registry.
        Distillation.unit_no += 1
        if self.object_id is None:
            self.object_id = f"D{Distillation.unit_no}"

        self.after: Dict[str, str] = {"bott": None, "dist": None}

        super().__post_init__()

    @validate_arguments
    def __call__(self, feed_stream: MaterialStream, **kwargs) -> Tuple[MaterialStream, MaterialStream]:

        for key, value in kwargs.items():
            if key not in self.__dataclass_fields__:
                raise AttributeError(f"{key} not recognized.")
            setattr(self, key, value)

        if feed_stream.data is not None:

            if self.check_data:
                with timer(verbose=self.verbose, operation="data check", unit=self.object_id) as _:
                    self._check_feed_stage()
                    self._check_num_cols()
                    self._check_units()

            with timer(verbose=self.verbose, operation="data prep", unit=self.object_id) as _:
                self.unit_data = self._format_unit_data()
                self.data = self._combine_unit_and_stream_data(feed_stream)

            with timer(verbose=self.verbose, operation="ML", unit=self.object_id) as _:
                bott_stream, dist_stream = self._predict(feed_stream)
        else:
            bott_stream = MaterialStream(stream_type="bott", before=self.object_id)
            dist_stream = MaterialStream(stream_type="dist", before=self.object_id)

        self._add_to_graph(feed_stream, "distillation")
        self._update_connections(input_stream=feed_stream, output_streams={"bott": bott_stream, "dist": dist_stream})

        return bott_stream, dist_stream

    def _predict(self, feed_stream: MaterialStream) -> Tuple[MaterialStream, MaterialStream]:

        model_fname = get_model(module=distillation_models, pattern=f"distillation_{len(feed_stream._suffixes)}_")

        with importlib.resources.path(distillation_models, model_fname) as model_path:
            with open(model_path, "rb") as model_file:
                stat_model = pickle.load(model_file)
                flow_model = pickle.load(model_file)
                duty_model = pickle.load(model_file)
                temp_model = pickle.load(model_file)

        # Prepare data
        if self.fillna:
            x = self.data.fillna(self.na_value).select_dtypes(exclude="object")
        else:
            x = self.data.select_dtypes(exclude="object")

        # Get predictions
        stat = stat_model.predict(x)
        flow = flow_model.predict(x)
        duty = duty_model.predict(x)
        temp = temp_model.predict(x)

        # Get flowrates
        bott_flow_perc = relu(flow, max_value=1.0)
        bott_flow = feed_stream.flow * bott_flow_perc
        dist_flow = feed_stream.flow - bott_flow

        # Turn flowrates into dictionaries with correct keys.
        col_dict = dict(
            zip(
                feed_stream._flow_columns,
                [f"flowrate_{suffix}" for suffix in feed_stream._suffixes],
            )
        )
        bott_flow = bott_flow.rename(columns=col_dict).to_dict("list")
        dist_flow = dist_flow.rename(columns=col_dict).to_dict("list")

        # Set heat & status for column.
        self.condensor_duty = (duty[:, 0] * feed_stream.flow_sum).tolist()
        self.reboiler_duty = (duty[:, 1] * feed_stream.flow_sum).tolist()
        if feed_stream.status is not None:
            self.status = stat * feed_stream.status
        else:
            self.status = stat

        # Prepare bottom & distillate stream data.
        bott_data = {
            "vapor_fraction": 0,
            "temperature": temp[:, 1].tolist(),
            "molecules": feed_stream.data[feed_stream._name_columns].to_dict("list"),
            "flowrates": bott_flow,
        }

        dist_data = {
            "vapor_fraction": 0,
            "temperature": temp[:, 0].tolist(),
            "molecules": feed_stream.data[feed_stream._name_columns].to_dict("list"),
            "flowrates": dist_flow,
        }

        if None not in self.after.values():
            bott_stream, dist_stream = update_material_streams(streams={self.after["bott"]: bott_data, self.after["dist"]: dist_data})

        else:
            bott_stream = MaterialStream(stream_type="bott")(**bott_data, before=self.object_id)
            dist_stream = MaterialStream(stream_type="dist")(**dist_data, before=self.object_id)

        bott_stream.status = self.status
        dist_stream.status = self.status

        return bott_stream, dist_stream

    def _check_feed_stage(self):
        """Make sure feed stage exists and makes sense.

        Raises:
            ValueError: if feed stage larger than # of stages.
        """

        if self.feed_stage is None:
            warnings.warn("feed_stage not specified, setting it equal to half no_stages.", stacklevel=2)
            if isinstance(self.no_stages, list):
                self.feed_stage = list((np.array(self.no_stages) / 2).round())
            elif isinstance(self.no_stages, int):
                self.feed_stage = round(self.no_stages / 2)

        if isinstance(self.feed_stage, int) and isinstance(self.no_stages, int):
            if not 1 < self.feed_stage < self.no_stages + 1:
                raise ValueError("Feed stage must be between 1 and no_stages + 1.")
        else:
            if not all(np.array(self.feed_stage) <= np.array(self.no_stages) + 1):
                raise ValueError("Feed stage can't be lager than number of stages.")
            if not all(1 <= np.array(self.feed_stage)):
                raise ValueError("Feed stage must be >= 1.")

    def _check_num_cols(self):
        """Operating options must not be redundantly/ambiguously specified."""

        unit_num_cols = []
        for key in self.__dataclass_fields__:
            if "grp" in self.__dataclass_fields__[key].metadata:
                group = self.__dataclass_fields__[key].metadata["grp"]
                if group == "num":
                    unit_num_cols.append(key)

        if None in [self.no_stages, self.pressure]:
            raise ValueError("Please specify # of stages & pressure.")

        operating_options = list({col: getattr(self, col) for col in unit_num_cols}.values())
        if sum(np.array(operating_options, dtype="object") != None) != 5:
            raise ValueError("Please specify exactly 2 out of operating options.")

        if self.reboiler_duty is not None or self.condensor_duty is not None:
            raise ValueError("Distillation doesn't support inputting reboiler/condensor duty currently.")



"""Base class for streams"""

from abc import ABC, abstractmethod
from typing import Any, ClassVar

import pandas as pd
from pydantic.dataclasses import dataclass

from ml4pd import registry


@dataclass
class Stream(ABC):
    """Base class for streams."""

    unit_no: ClassVar[int] = -1
    base_type: ClassVar[str] = "stream"
    object_id: str = None
    before: str = None
    after: str = None
    check_data: bool = True
    verbose: bool = False

    def __post_init__(self):

        self.data: pd.DataFrame = None
        self.status: pd.DataFrame = None

        registry.add_element(self)

    @abstractmethod
    def __call__(self, *args: Any, **kwds: Any) -> Any:
        """Gather input into df for ML."""

    def __eq__(self, other: object) -> bool:
        return all((self.data.to_numpy() == other.data.to_numpy()).flatten())



"""This module contains some useful functions to both user and internal code."""

import time
from contextlib import contextmanager


@contextmanager
def timer(verbose: bool, operation: str, unit: str):
    """Use this function with 'with' to time any block of code."""

    try:
        start_time = time.time()
        yield start_time
    finally:
        if verbose:
            print(f"{unit}: {operation} took {round(time.time() - start_time, 2)} seconds.")



import itertools
import random
from abc import ABC, abstractmethod
from typing import List, Literal, Union

import numpy as np
import pandas as pd
from pydantic import validate_arguments
from pydantic.dataclasses import dataclass


@dataclass
class Dimension(ABC):
    """
    Base class for objects used to generate data samples.

    ## Parameters:
    - samples: The # of rows of generated list.
    - group: The # of columns of generated list.
    - sampling: sampling scheme. Depends on type of dimension.

    """

    samples: int = None
    sampling: str = None
    group: int = None

    @abstractmethod
    def generate(self):
        """Generate one/multiple values depending on samples."""
        pass


@dataclass
class Integer(Dimension):
    """Generates one / a list of integers based on specified sampling method."""

    min_value: int = None
    max_value: int = None
    samples: int = None
    sampling: Literal["random-uniform", "linear"] = "linear"

    def __post_init__(self):
        self.values: List[int] = None

    def generate(self):
        if self.sampling == "random-uniform":
            values = self.generate_random_uniform_values()
        elif self.sampling == "linear":
            values = self.generate_linear_values()

        if len(values) == 1:
            values = values.round(5)[0]
        else:
            values = values.round(5).tolist()
        return values

    def generate_random_uniform_values(self):
        return np.random.randint(low=self.min_value, high=self.max_value + 1, size=self.samples)

    def generate_linear_values(self):
        return np.linspace(start=self.min_value, stop=self.max_value, num=self.samples, dtype=int)


@dataclass
class Float(Dimension):

    min_value: Union[float, int] = None
    max_value: Union[float, int] = None
    samples: int = None
    sampling: Literal["random-uniform", "linear", "log"] = "linear"

    def generate(self):
        if self.sampling == "random-uniform":
            values = self.generate_random_uniform_values()
        elif self.sampling == "linear":
            values = self.generate_linear_values()
        elif self.sampling == "log":
            values = self.generate_log_values()

        if len(values) == 1:
            values = values.round(5)[0]
        else:
            values = values.round(5).tolist()
        return values

    def generate_random_uniform_values(self):
        return np.random.uniform(low=self.min_value, high=self.max_value, size=self.samples)

    def generate_linear_values(self):
        return np.linspace(start=self.min_value, stop=self.max_value, num=self.samples, dtype=float)

    def generate_log_values(self):
        return np.logspace(np.log10(self.min_value), np.log10(self.max_value), self.samples, dtype=float)


@dataclass
class Choice(Dimension):

    choices: list = None
    group: int = 1
    samples: int = None

    def generate(self):
        values = []
        for _ in range(self.samples):
            values.append(random.sample(self.choices, self.group))

        if self.group == 1:
            values = np.array(values).flatten().tolist()

        return values


@validate_arguments
def generate_grid(**kwargs: Union[Dimension, list, float, int, str]) -> pd.DataFrame:

    search_space = {}
    dtypes = []
    for key, range in kwargs.items():
        if isinstance(range, Dimension):
            search_space[key] = range.generate()
        elif isinstance(range, list):
            search_space[key] = range
        else:
            search_space[key] = [range]
        dtypes.append((key, object))

    # Turn into nested list
    search_space = list(search_space.values())

    # Created permutations
    search_space = np.array(list(itertools.product(*search_space)), dtype=dtypes)

    # Turn into df
    df = pd.DataFrame(search_space).infer_objects()

    return df



"""
Module contains class Registry and a registry object that
keeps track of all initialized objects. To be used by flowsheet
to plot & pass data through graph structure.
"""

__all__ = ["registry"]


class Registry:
    """
    Class used to keep track of all initialized objects.

    Attributes:
        all_streams (dict): dictionary of all streams initialized.
        all_columns (dict): dicitonary of all columns initiliazed.

    """

    def __init__(self):
        self.all_streams = {}
        self.all_columns = {}

    def add_element(self, element):
        """Add element to all streams/columns based on its type."""
        if element.base_type == "stream":
            if element.object_id in self.all_streams:
                raise KeyError(f"{element.object_id} already taken.")
            self.all_streams[element.object_id] = element
        elif element.base_type == "unit-op":
            if element.object_id in self.all_columns:
                raise KeyError(f"{element.object_id} already taken.")
            self.all_columns[element.object_id] = element

    def remove_element(self, object_id: str):
        """Remove item given its id from all streams/columns."""
        if object_id in self.all_streams:
            del self.all_streams[object_id]
        elif object_id in self.all_columns:
            del self.all_columns[object_id]
        else:
            raise KeyError("id not found.")

    def clear_data(self):
        """Remove all streams & columns from registry."""
        self.all_streams = {}
        self.all_columns = {}

    def get_element(self, object_id: str):
        """Return object that corresponds to the id."""
        if object_id in self.all_streams:
            return_item = self.all_streams[object_id]
        elif object_id in self.all_columns:
            return_item = self.all_columns[object_id]
        else:
            raise KeyError("id not found.")

        return return_item

    def get_all_streams(self):
        """Returns dictionary of all streams."""
        return self.all_streams

    def get_all_columns(self):
        """Returns dictionary of all columns."""
        return self.all_columns


registry = Registry()



from setuptools import setup, find_packages
import pathlib
from datetime import date


here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")
today = date.today().strftime(r"%Y.%m.%d")

requirements = (here / "requirements.txt").read_text(encoding="utf-8").split("\n")
test = [
    "seaborn",
    "ipykernel",
    "ipython",
    "pytest",
    "black",
    "pylint",
    "tqdm",
    "mkdocs-material",
    "mkdocs-jupyter",
]

setup(
    name="ml4pd",
    version=today,
    maintainer="Hien Vo",
    maintainer_email="hvo@nrel.gov",
    description="ML4PD - an open-source libray for building Aspen-like process models via machine learning.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=["Programming Language :: Python :: 3.7", "Programming Language :: Python :: 3.8"],
    packages=find_packages(exclude=["tests"]),
    python_requires=">=3.8",
    install_requires=requirements,
    include_package_data=True,
    extras_require={"test": test},
)



"""Base class for unit ops."""

import warnings
from abc import ABC, abstractmethod
from dataclasses import field
from typing import Any, ClassVar, Dict, List

import networkx
import numpy as np
import pandas as pd
from pint import UnitRegistry
from pydantic.dataclasses import dataclass

from ml4pd import registry
from ml4pd.streams.stream import Stream

ureg = UnitRegistry()


@dataclass
class UnitOp(ABC):
    """Base class for aspen unit ops."""

    unit_no: ClassVar[int] = -1
    base_type: ClassVar[str] = "unit-op"
    object_id: str = None
    before: Any = field(default_factory=list)
    after: Any = field(default_factory=dict)
    check_data: bool = True
    verbose: bool = False
    fillna: bool = True
    na_value: float = 0

    def __post_init__(self):

        self.status: np.ndarray = None
        self.unit_data: pd.DataFrame = None
        self.data: pd.DataFrame = None
        self.networkx: networkx.DiGraph = None
        self._quant_with_units: Dict[str, Dict[str, List[str]]] = {}

        # Fill up quants with units:
        for key in self.__dataclass_fields__:
            if "_units" in key:
                self._quant_with_units[key] = {self.__dataclass_fields__[key].default: []}

        for key in self.__dataclass_fields__:
            if "grp" in self.__dataclass_fields__[key].metadata:
                for pattern in ["rate", "duty", "pres", "temp"]:
                    if pattern in key:
                        unit = list(self._quant_with_units[f"{pattern}_units"].keys())[0]
                        self._quant_with_units[f"{pattern}_units"][unit].append(key)

        registry.add_element(self)

    @abstractmethod
    def __call__(self, **kwds: Any) -> Any:
        """Combine streams & unit data, then call _predict."""

    @abstractmethod
    def _predict(self, **kwds: Any) -> Any:
        """Produce unit op results and create product streams."""

    def __eq__(self, other: object) -> bool:
        """Compare unit ops based on its ML feature."""
        return all((self.unit_data.to_numpy() == other.unit_data.to_numpy()).flatten())

    def _update_connections(self, input_stream: Stream, output_streams: Dict[str, Stream]):
        """Add unit's object_id to input & output streams."""

        if input_stream.object_id not in self.before:
            self.before.append(input_stream.object_id)
            input_stream.after = self.object_id

        for stream_type, stream in output_streams.items():
            self.after[stream_type] = stream.object_id

    def _add_to_graph(self, feed_stream: Stream, unit_type: str):
        """Update networkx with feed stream and unit."""

        if feed_stream.before is None and len(self.before) == 0:
            self.networkx = networkx.DiGraph()

        if feed_stream.before is None:
            self.networkx.add_node(f"{feed_stream.object_id}_input")
            self.networkx.add_edge(f"{feed_stream.object_id}_input", self.object_id, label=feed_stream.object_id)
        else:
            self.networkx = getattr(registry.get_element(feed_stream.before), "networkx")
            self.networkx.add_edge(feed_stream.before, self.object_id, label=feed_stream.object_id)

        self.networkx.add_node(self.object_id, type=unit_type)

    def _check_units(self):
        """
        Loop through _quant_with_units dict and convert user specified quantities
        to default units to conform to training data. Will throw an error for
        more 'exotic' units. In that case, user will have to manually convert units.
        """

        for unit_type, default_and_quants in self._quant_with_units.items():
            default_unit = list(default_and_quants.keys())[0]
            user_unit = getattr(self, unit_type)
            if user_unit != default_unit:
                quants = default_and_quants[default_unit]
                for quant in quants:
                    user_quant = getattr(self, quant)
                    if user_quant is not None:
                        default_quant = ureg.Quantity(user_quant, ureg(user_unit)).to(default_unit).magnitude
                        if isinstance(default_quant, np.ndarray):
                            setattr(self, quant, default_quant.tolist())
                        else:
                            setattr(self, quant, default_quant)
                setattr(self, unit_type, default_unit)

    def _format_unit_data(self) -> pd.DataFrame:
        """Collect unit data into dataframe."""

        unit_data = {}
        for key in self.__dataclass_fields__:
            if "grp" in self.__dataclass_fields__[key].metadata:
                group = self.__dataclass_fields__[key].metadata["grp"]
                if group == "str":
                    unit_data[f"{self.object_id}_{key}"] = self.__annotations__[key].__args__.index(getattr(self, key))
                elif group == "num":
                    unit_data[f"{self.object_id}_{key}"] = getattr(self, key)

        try:
            unit_data = pd.DataFrame(unit_data)
        except ValueError:
            unit_data = pd.DataFrame(unit_data, index=[0])

        return unit_data

    def _combine_unit_and_stream_data(self, feed) -> pd.DataFrame:
        """Concat feed & unit data. If they are not of compatible length, raise error."""

        if len(self.unit_data) == len(feed.data):
            data = pd.concat([self.unit_data, feed.data], axis=1)

        elif len(self.unit_data) == 1 and len(feed.data) != 1:
            warnings.warn("Stream and unit data shapes don't exactly match so broadcasting will be used.", stacklevel=2)
            data = pd.concat([self.unit_data.copy(deep=True)] * len(feed.data), ignore_index=True)
            data = pd.concat([data, feed.data], axis=1)

        elif len(feed.data) == 1 and len(self.unit_data) != 1:
            warnings.warn("Stream and unit data shapes don't exactly match so broadcasting will be used.", stacklevel=2)
            data = pd.concat([feed.data.copy(deep=True)] * len(self.unit_data), ignore_index=True)
            data = pd.concat([data, self.unit_data], axis=1)

        else:
            raise ValueError("Data shape doesn't match input stream.")

        return data

    def clear_data(self):
        """Clear user-defined data."""

        # Delete ml input data
        for key in self.__dataclass_fields__:
            if "grp" in self.__dataclass_fields__[key].metadata:
                delattr(self, key)

        # Delete other data
        delattr(self, "status")
        delattr(self, "unit_data")
        delattr(self, "data")
